/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// Author: Joe Romano

#include "ros/ros.h"

#include <actionlib/server/action_server.h>
#include <slipgrip_controller/GripperSlipServoAction.h>
#include <slipgrip_controller/GripperSlipServoData.h>
#include <slipgrip_controller/GripperSlipServoCommand.h>
#include <std_srvs/Empty.h>

class Pr2GripperSlipServo
{
private:
  typedef actionlib::ActionServer<slipgrip_controller::GripperSlipServoAction> GAS;
  typedef GAS::GoalHandle GoalHandle;
public:
  Pr2GripperSlipServo(ros::NodeHandle &n) :
    node_(n),
    action_server_(node_, "slip_servo",
                   boost::bind(&Pr2GripperSlipServo::goalCB, this, _1),
                   boost::bind(&Pr2GripperSlipServo::cancelCB, this, _1)),
    has_active_goal_(false)
  {
    ros::NodeHandle pn("~");

    // param server stuff?
    //pn.param("goal_threshold", goal_threshold_, 0.01);
    //pn.param("stall_velocity_threshold", stall_velocity_threshold_, 1e-6);
    //pn.param("stall_timeout", stall_timeout_, 0.1);

    // we will listen for messages on "state" and send messages on "find_contact"
    pub_controller_command_ =
      node_.advertise<slipgrip_controller::GripperSlipServoCommand>("slip_servo", 1);
    sub_controller_state_ =
      node_.subscribe("slip_servo_state", 1, &Pr2GripperSlipServo::controllerStateCB, this);

    watchdog_timer_ = node_.createTimer(ros::Duration(1.0), &Pr2GripperSlipServo::watchdog, this);

  }

  ~Pr2GripperSlipServo()
  {
    pub_controller_command_.shutdown();
    sub_controller_state_.shutdown();
    watchdog_timer_.stop();
  }

private:

  ros::NodeHandle node_;
  GAS action_server_;
  ros::Publisher pub_controller_command_;
  ros::Subscriber sub_controller_state_;
  ros::Timer watchdog_timer_;

  bool has_active_goal_;
  bool firstDrop;
  GoalHandle active_goal_;
  ros::Time goal_received_;

  double goal_threshold_;
  double stall_velocity_threshold_;
  double stall_timeout_;
  ros::Time last_movement_time_;


  void watchdog(const ros::TimerEvent &e)
  {
    ros::Time now = ros::Time::now();

    // Aborts the active goal if the controller does not appear to be active.
    if (has_active_goal_)
    {
      bool should_abort = false;
      if (!last_controller_state_)
      {
        should_abort = true;
        ROS_WARN("Aborting goal because we have never heard a controller state message.");
      }
      else if ((now - last_controller_state_->header.stamp) > ros::Duration(5.0))
      {
        should_abort = true;
        ROS_WARN("Aborting goal because we haven't heard from the controller in %.3lf seconds",
                 (now - last_controller_state_->header.stamp).toSec());
      }

      if (should_abort)
      {
        // Marks the current goal as aborted.
        active_goal_.setAborted();
        has_active_goal_ = false;
      }
    }
  }

  void goalCB(GoalHandle gh)
  {
    // Cancels the currently active goal.
    if (has_active_goal_)
    {
      // Marks the current goal as canceled.
      active_goal_.setCanceled();
      has_active_goal_ = false;
    }
 
    gh.setAccepted();
    active_goal_ = gh;
    has_active_goal_ = true;
    firstDrop = true;  // set our first drop flag
    goal_received_ = ros::Time::now();
    
    // Sends the command along to the controller.
    pub_controller_command_.publish(active_goal_.getGoal()->command);
    
    last_movement_time_ = ros::Time::now();
  }

  void cancelCB(GoalHandle gh)
  {
    
    if (active_goal_ == gh)
    {
      // Stops the controlle.
      if (last_controller_state_)
      {
        // stop the gripper
        slipgrip_controller::GripperSlipServoCommand cmd_send;
        cmd_send.deformation_limit = 0.0;
        cmd_send.fingertip_start_force = 0.0;
        cmd_send.fingertip_force_limit = 0.0;
        cmd_send.max_effort = 0.1;
        pub_controller_command_.publish(cmd_send);
      }

      // Marks the current goal as canceled.
      active_goal_.setCanceled();
      has_active_goal_ = false;
    }
  }


  slipgrip_controller::GripperSlipServoDataConstPtr last_controller_state_;
  void controllerStateCB(const slipgrip_controller::GripperSlipServoDataConstPtr &msg)
  {
    last_controller_state_ = msg;
    ros::Time now = ros::Time::now();

    if (!has_active_goal_)
      return;

    slipgrip_controller::GripperSlipServoGoal goal;
    goal.command = active_goal_.getGoal()->command;

    slipgrip_controller::GripperSlipServoFeedback feedback;
    feedback.data = *msg;

    slipgrip_controller::GripperSlipServoResult result;
    result.data = *msg;
    
    // if we are actually in a slip_servo control state
    if(feedback.data.realtime_controller_state == 6)
    {
      // if the object is dropped
      if ( feedback.data.dropped && firstDrop )
      {
        //active_goal_.setAborted(result);

        // tell the gripper to close gently 
        slipgrip_controller::GripperSlipServoCommand cmd_send;
        cmd_send.deformation_limit = -1.0;
        cmd_send.fingertip_start_force = 12.0;
        cmd_send.fingertip_force_limit = 15.0;
        cmd_send.max_effort = 30.0;
        pub_controller_command_.publish(cmd_send);
      
        firstDrop = false; // set our flag to false to indicate we already entered "dropped" mode

        //has_active_goal_ = false;
      }

      // if we somehow magically found the object again then return to our previous goal
      else if( !feedback.data.dropped && !firstDrop)
      {
        pub_controller_command_.publish(goal.command);
        firstDrop = true;
      }
    }
    else
      has_active_goal_ = false;
    
    active_goal_.publishFeedback(feedback);
  }
};


int main(int argc, char** argv)
{
  ros::init(argc, argv, "slip_servo_node");
  ros::NodeHandle node;
  Pr2GripperSlipServo jte(node);

  ros::spin();

  return 0;
}
