//@author Joe Romano
//@contact joeromano@gmail.com
//@date 06.12.10
//@brief - pressureObserver.h - helper class to watch the presure sensors on the robot gripper and perform important data storage and manipulation

#include <pr2_hardware_interface/hardware_interface.h>
#include <slipgrip_controller/GripperSensorData.h>
#include <slipgrip_controller/digitalFilter.h>

#ifndef _PRESSURE_OBSERVER
#define _PRESSURE_OBSERVER

#define UNITS_PER_N 1600

class pressureObserver{
  
  public:
    // functions
    pressureObserver(pr2_hardware_interface::PressureSensor* left,pr2_hardware_interface::PressureSensor* right);    // ctor
    ~pressureObserver();                                               // default destructor
    void spin();                                                       // method to update sensor data. should run at 1khz
    bool graspContact(int contactsDesired);                              // detects finger impacts
    void updateZeros2();                                               // method to update our zero value offsets (zero_offset)
    bool checkPlaceContact(double dF);                                         // method to look for impact slip and return true
  //    bool checkSlip(double centroid_motion_limit, double centroid_force_limit);
  bool checkSlip(double centroid_motion_limit = 0.008, double centroid_force_limit = 0.2);

    double padForce_cur;                                               // fingerpad force average between both fingers
    double padForce_cur_nonbiased;                                     // force average from the non-biased pressures
    double padForce_left_cur, padForce_left_prev;                      // current cumulative force on our left/rigth fingerpads
    double padForce_left_cur_nonbiased;
    double padForce_right_cur, padForce_right_prev;                    // previous cumulative force on our left/right fingerpads
    double padForce_right_cur_nonbiased;
    double totalLoading_left, totalLoading_right;                      // total dynamic laoding on our left/right pads
    double totalLoading,totalLoading_prev;                                               // left and right loading average
    double cellLoading_left, cellLoading_right;                        // cellular loading on the left/right fingers
    double forceDropped;
    double lfForce_left, lfForce_right;
    double dFdt, dFdt_avg;
    double x_left_bp_centroid, y_left_bp_centroid, x_right_bp_centroid, y_right_bp_centroid;
    //PerformanceCounter  counter;

    // storage of our pressure sensor state information. _now is the most recent raw data. _prev is the last subsampled value
    // _current is the current subsampled value
    slipgrip_controller::GripperSensorData pressure_now, pressure_prev, pressure_current;
    slipgrip_controller::GripperSensorData pressure_thresh_now, pressure_thresh_prev, pressure_thresh_current;
    slipgrip_controller::GripperSensorData pressure_prev_zerod, pressure_current_zerod;
    slipgrip_controller::GripperSensorData pressure_prev_bias, pressure_cur_bias;
    slipgrip_controller::GripperSensorData pressure_bp;

    digitalFilter *pressureLFFilt_left[22];
    digitalFilter *pressureLFFilt_right[22];
    digitalFilter *centroid_bp_filt[4];

    bool left_contact, right_contact;                                  // flags to indicate whether a left or right contact happened
    bool placeContact;

  private:
    void getPadCentroid(boost::array<double,22> pressureArray, double *x, double *y);
    void updateXYBPCentroids();
    bool updatePressureState();                                        // method to acquire new data from both fingers and apply appropriate subsampling 
    void updateTotalLoading();                                         // method to update the toal loading of each finger totalLoad_left _right
    void updateCellLoading();                                          // method to update the cell-level loading of each finger cellLoad_left _right
    void zero();                                                       // method to zero out data
    void bias_estimator(double zero_range);                            // remove mean bias for near zero data
    double getPadForce(boost::array<double,22> pressureArray);         // translate fingertip pressure vector to cumultive sum on the fingerpad (just the pad)
    void updateContactState();                                         // update the state of the robots contact sensors


    pr2_hardware_interface::PressureSensor* left_finger;               // a copy of our pressure sensor handle
    pr2_hardware_interface::PressureSensor* right_finger;              // a copy of our pressure sensor handle

    int subSample_cnt;                                                 // a counter to keep track of how many samples we ignored during subsampling
    int subSample_steps;                                               // 1+ the number of values to skip during subsampling
    int zero_samples;                                                  // the number of samples to use when trying to evaluate our zero offset
    int zero_cnt;                                                      // counter to keep track of how many samples we've stored for zero estimation
    slipgrip_controller::GripperSensorData threshold_max;              // a temporary storage container for our max threshold values
    slipgrip_controller::GripperSensorData threshold_min;              // a temporary storage container for our min threshold values
    slipgrip_controller::GripperSensorData thresholds;                 // a storage container for our threshold values
    slipgrip_controller::GripperSensorData zero_offset;                // a storage container for our zero offset values

    double time_prev, time_cur;                                        // stroage containers for the previous and current pressure update times
    double dt;                                                         // the time expired between pressure update iterations

    int debug_count;                                                   // debug loop counter
};



#endif // _PRESSURE_OBSERVER
