/*
**   rtstorage.h
**
**   Class control::RTStorage<MessageType>
**
**   This class provides realtime-safe storage as well as services to
**   record and upload data.
**
**
**   Provides:
**	RTStorage(std::string myPublishName)
**	~RTStorage()
**
**	bool init(ros::NodeHandle &n)
**	bool init(ros::NodeHandle &n, int samples)
**
**		Initialize the storage buffer, services, and
**		publisher.  The first version reads the number of
**		samples from the parameter server.
**
**	unsigned int size()	Report the size of the buffer.
**
**   #	void start()		Start recording.
**   #	void stop()		Start recording.
**	void wait()		Start recording.
**	void upload()		Upload after recording is complete
**	void collect()		Start recording, wait, then upload
**
**   #	void store(const MessageType &data)
**
**		Store a single sample
**
**   where # denotes REALTIME functions
**
**
**   Services:
**	myPublishName/start	Start recording
**	myPublishName/stop	Stop recording
**	myPublishName/wait	Wait for recording to be complete
**	myPublishName/upload	Upload the recorded data
**	myPublishName/collect	Start, wait, then upload
**
**
**   Topics:
**	myPublishName/data	Recorded data
*/

#ifndef  CONTROL_RTSTORAGE_H
#define  CONTROL_RTSTORAGE_H

#include <ros/ros.h>
#include <std_srvs/Empty.h>


namespace control
{
  template<class MessageType>
  class RTStorage
    {
    private:
      // Buffer space
      std::vector<MessageType> buffer_;

      // the name we'd like to publish our message under
      std::string publishedName;

      // Storage status.  This simply determines whether data should
      // be stored.  Marked volatile so that the non-realtime thread can
      // watch its value change.
      volatile bool  storing_;

      // Data index.  This is the number of completed entries in the buffer.
      // The realtime thread increments this, after it stores data.  The
      // non-realtime thread can reset it to zero to restart the storage
      // and can examine its value.  We hence mark it volatile.
      volatile unsigned int  completed_;

      // The services
      ros::ServiceServer start_srv_;	// To start storage
      ros::ServiceServer stop_srv_;	// To stop storage
      ros::ServiceServer wait_srv_;	// To wait until the data is complete
      ros::ServiceServer upload_srv_;	// To upload data (without storage)
      ros::ServiceServer collect_srv_;	// To store, wait, and upload data

      // The publisher
      ros::Publisher storage_pub_;	// Output stored data


    public:
      /// default ctor with optional name for data node
      RTStorage()
	{
	  // Mark the buffer as non-existant (empty) and not storing. 
	  completed_ = 0;
	  storing_   = false;
	}


      ~RTStorage()
	{
	}



      /// Initialization.
      bool init(ros::NodeHandle &n, std::string myPublishName = "/storage")
      {
	// Try to read the storage buffer size from the parameter server.
	int samples;

	if (!n.getParam("samples", samples))
	  {
	    samples = 0;
	    ROS_WARN("No parameter 'samples' in namespace '%s' - not storing...",
		     n.getNamespace().c_str());
	  }

	return init(n, samples,myPublishName);
      }

      bool init(ros::NodeHandle &n, int samples, std::string myPublishName = "/storage")
      {
	// Allocate the storage buffer space.  Note we allocate an additional
	// one hundred entries.  The publication doesn't flush all of it's
	// internal buffers (it appears) so the last samples get stuck in the
	// output queue...
	if (samples > 0)
	  {
	    samples = samples + 100;
	    buffer_.resize(samples);
	  }
	else
	  {
	    samples = 0;
	    buffer_.clear();
	  }
	samples    = buffer_.size();
	completed_ = 0;
	storing_   = false;

	// Create the services and publisher.
	start_srv_   = n.advertiseService(myPublishName+"/start",   &RTStorage::startService,   this);
	stop_srv_    = n.advertiseService(myPublishName+"/stop",    &RTStorage::stopService,    this);
	wait_srv_    = n.advertiseService(myPublishName+"/wait",    &RTStorage::waitService,    this);
	upload_srv_  = n.advertiseService(myPublishName+"/upload",  &RTStorage::uploadService,  this);
	collect_srv_ = n.advertiseService(myPublishName+"/collect", &RTStorage::collectService, this);

	storage_pub_ = n.advertise<MessageType>(myPublishName+"/data", (samples>0?samples:1));

	return true;
      }



      /// Report the size of the buffer.
      unsigned int size()
      {
	return buffer_.size();
      }



      /// REALTIME: Start/stop storage.  Note starting also resets the buffer to the beginning.
      void start()
      {
	// Mark the buffer empty and storing.
	completed_ = 0;
	storing_   = true;
      }

      void stop()
      {
	// Simply mark as not storing.
	storing_ = false;
      }



      /// REALTIME: Store a sample.
      void store(const MessageType &data)
      {
	// Return if we aren't actually storing.
	if (!storing_)
	  return;

	// Only save the data if there is room in the buffer.  Note we
	// rely on the fact that the other functions are all
	// non-realtime, which can't interrupt this realtime function.
	// Also they only reset the completed count to zero and hence
	// can't violate the condition.
	if (completed_ < buffer_.size())
	  {
	    buffer_[completed_] = data;
	    completed_++;
	  }

	// Stop storing if all buffer elements are complete.
	if (completed_ >= buffer_.size())
	  {
	    storing_ = false;
	  }
      }



      /// Wait, upload, collect.
      void collect()
      {
	start();
	wait();
	upload();
      }

      void wait()
      {
	ros::Time started;		// Starting time
	double    max_seconds;		// Max allowed time

	// Record the current time and predict a reasonable expected duration.
	started     = ros::Time::now();
	max_seconds = 2.0 + 0.001 * buffer_.size();

	// Now wait until the buffer is full.
	// ROS_INFO("Waiting... at most %f seconds", max_seconds);
	while (storing_)
	  {
	    // Sleep for 1ms as not to hog the CPU.
	    ros::Duration(0.001).sleep();

	    // Make sure we don't hang here forever.
	    if (ros::Time::now() - started > ros::Duration(max_seconds))
	      {
		ROS_WARN("Waiting for buffer to fill aborted after %f seconds!",
			 max_seconds);
		break;
	      }
	  }
      }

      void upload()
      {
	// Publish the current buffer contents.  Note we rely on the fact that
	// the storage will only increment the number of completed entries.
	unsigned int i;
	for (i = 0 ; i < completed_ ; i++)
	  {
	    storage_pub_.publish(buffer_[i]);
	  }
      }


    private:
      /// Services (redirect)
      bool startService(std_srvs::Empty::Request& req,
			std_srvs::Empty::Response& resp)
      {
	start();
	return true;
      }

      bool stopService(std_srvs::Empty::Request& req,
		       std_srvs::Empty::Response& resp)
      {
	stop();
	return true;
      }

      bool waitService(std_srvs::Empty::Request& req,
		       std_srvs::Empty::Response& resp)
      {
	wait();
	return true;
      }

      bool uploadService(std_srvs::Empty::Request& req,
			 std_srvs::Empty::Response& resp)
      {
	upload();
	return true;
      }

      bool collectService(std_srvs::Empty::Request& req,
			  std_srvs::Empty::Response& resp)
      {
	collect();
	return true;
      }
    };
}

#endif
