%{
*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  \author Joe Romano
 *********************************************************************/
%}
%
%  @author Joe Romano
%  @brief Function to plot debugging data during slipgrip_controller trials
%  composed of accelerations, fingertip pressures (left/right), and joint
%  information (velocity,position,effort)
%
%  Usage - showAllSignals('myTrial');
%  Note: You only need to include the base name: myTrial, and not the
%  sub-names such as myTrial_acc.rtp ...
%
function [] = showAllSignals(myfilename)

[pRaw_left pRaw_right pRawTime pFilt_left pFilt_right pFiltTime acc_combined accTime joint_pos joint_vel joint_effort joint_time ] = loadAllRTP(myfilename);
UNITS_PER_N = 1600;

% subsample the pressure like the software would
[pRaw_left_sub pRawTime_sub] = subsamplePressureVector(pRaw_left, pRawTime);
[pRaw_right_sub pRawTime_sub] = subsamplePressureVector(pRaw_right, pRawTime);
[pFilt_left_sub pFiltTime_sub] = subsamplePressureVector(pFilt_left, pFiltTime);
[pFilt_right_sub pFiltTime_sub] = subsamplePressureVector(pFilt_right, pFiltTime);

% convert to Newtons
pFilt_left_sub = pFilt_left_sub/UNITS_PER_N;
pFilt_right_sub = pFilt_right_sub/UNITS_PER_N;

% start our time vectors from 0
toffset = pRawTime_sub(1);
pRawTime_sub = pRawTime_sub - toffset;
pFiltTime_sub = pFiltTime_sub - toffset;
accTime = accTime - toffset;
joint_time = joint_time - toffset;
tstart = pRawTime_sub(1);
tend = pRawTime_sub(end);


% create our first figure of 5 subplots

subplot(5,1,1);
plot(joint_time,joint_pos,'b');
ylabel('Joint Position (m)');
xlim([ tstart tend ] );


subplot(5,1,2);
[b,a] = butter(1,5/(500));
joint_effort_filt = filter(b,a,joint_effort);
plot(joint_time,joint_effort,joint_time,joint_effort_filt);
ylabel('Joint Effort (N)');
legend('Raw Effort', '5-Hz LP Filtered');
xlim([ tstart tend ] );



subplot(5,1,3);
[acc_combined_filt] = hpAccelerationMatrix(acc_combined,3000,60);
plot(accTime,acc_combined,accTime,acc_combined_filt);
ylabel('Raw Acceleration (m/s^2)');
legend('X','Y','Z','X_filt','Y_filt','Z_filt');
xlim([ tstart tend ] );


subplot(5,1,4);
plot(pRawTime_sub,pRaw_left_sub,pRawTime_sub,pRaw_right_sub);
ylabel('Subsampled Raw Pressure (N) - Left/Right Fingers');
xlim([tstart tend] );

subplot(5,1,5);
plot(pFiltTime_sub,pFilt_left_sub,pFiltTime_sub,pFilt_right_sub);
ylabel('Subsampled Filtered Pressure (N) - Left/Right Fingers');
xlim([ tstart tend ] );


%create our second figure showing some processing
%{
figure();

subplot(6,1,1);
plot(joint_time,joint_pos,'b');
ylabel('joint position');
%hold on;
%plot(joint_time,joint_vel,'r');
xlim([ tstart tend ] );

subplot(6,1,2);
%left_force = sumPressureVectorFingerpad(pFilt_left_sub);
%right_force = sumPressureVectorFingerpad(pFilt_right_sub);
left_force = sumPressureVectorFinger(pFilt_left_sub);
right_force = sumPressureVectorFinger(pFilt_right_sub);
[b,a] = butter(1,5/((1/.041)/2));
left_force_filt = filter(b,a,left_force);
right_force_filt = filter(b,a,right_force);
plot(pFiltTime_sub,(left_force+right_force)/2,pFiltTime_sub,left_force,pFiltTime_sub,right_force);
hold on;
plot(pFiltTime_sub,(left_force_filt+right_force_filt)/2,pFiltTime_sub,left_force_filt,pFiltTime_sub,right_force_filt);
ylabel('grip force');
xlim([tstart tend]);

subplot(6,1,3);
gfd_l = [0; diff(left_force)./diff(pFiltTime_sub)'];
gfd_r = [0; diff(right_force)./diff(pFiltTime_sub)'];
plot(pFiltTime_sub,gfd_l, pFiltTime_sub,gfd_r, pFiltTime_sub, (gfd_r+gfd_l)/2);
ylabel('grip force deriv');
xlim([tstart tend]);

subplot(6,1,4);
%cd_l = cellPadDerivSum(pFilt_left_sub,pFiltTime_sub);
%cd_r = cellPadDerivSum(pFilt_right_sub,pFiltTime_sub);cd_l =
cd_l = cellFingerDerivSum(pFilt_left_sub,pFiltTime_sub);
cd_r = cellFingerDerivSum(pFilt_right_sub,pFiltTime_sub);
plot(pFiltTime_sub,cd_l, pFiltTime_sub,cd_r, pFiltTime_sub, (cd_l+cd_r)/2);
ylabel('cell deriv');
xlim([tstart tend]);

subplot(6,1,5);
%plot(pFiltTime_sub,cd_l-abs(gfd_l), pFiltTime_sub,cd_r-abs(gfd_r), pFiltTime_sub,((cd_l+cd_r)/2) - ((abs(gfd_r)+abs(gfd_l))/2));
plot(pFiltTime_sub,((cd_l+cd_r)/2) - ((abs(gfd_r)+abs(gfd_l))/2));
ylabel('4-abs(3) above');
xlim([tstart tend]);

subplot(6,1,6);
left_slip = slipDeriv(pFilt_left_sub,pFiltTime_sub);
right_slip = slipDeriv(pFilt_right_sub,pFiltTime_sub);
[b,a] = butter(1, 4/25,'low');               % test out a 1st order l.p. filter for the heck of it 
combined_slip_filt = filter(b,a,(left_slip+right_slip)/2);
plot(pFiltTime_sub,(left_slip+right_slip)/2,pFiltTime_sub,left_slip,pFiltTime_sub,right_slip,pFiltTime_sub,combined_slip_filt);
ylabel('slip');
xlim([tstart tend]);
%}